from dataclasses import dataclass
import inspect
import logging
from typing import (
    Annotated,
    Any,
    Callable,
    Literal,
    Sequence,
    TypeVar,
    cast,
    get_args,
    get_origin,
    get_type_hints
)

import annotated_types
from pydantic.fields import FieldInfo

import gradio as gr


LOGGER = logging.getLogger(__name__)


T = TypeVar('T')


class UnsupportedTypeError(ValueError):
    def __init__(self, type_hint: Any) -> None:
        super().__init__(f'Unsupported type: {type_hint}')
        self.type_hint = type_hint


@dataclass(frozen=True)
class ParsedFieldInfo[T]:
    title: str | None = None
    ge: annotated_types.SupportsGe | None = None
    le: annotated_types.SupportsLe | None = None
    valid_values: Sequence[T] | None = None


DEFAULT_PARSED_FIELD_INFO = ParsedFieldInfo[Any]()


def get_literal_values(literal) -> tuple:
    return get_args(literal)


def parse_pydantic_field_info(
    field_info: FieldInfo,
    valid_values: Sequence[T] | None = None
) -> ParsedFieldInfo:
    ge: annotated_types.SupportsGe | None = None
    le: annotated_types.SupportsLe | None = None
    for meta in field_info.metadata:
        if isinstance(meta, annotated_types.Ge):
            ge = meta.ge
        if isinstance(meta, annotated_types.Le):
            le = meta.le
    return ParsedFieldInfo[T](
        title=field_info.title,
        ge=ge,
        le=le,
        valid_values=valid_values
    )


def parse_pydantic_field(type_hint) -> tuple[Any, ParsedFieldInfo]:
    origin = get_origin(type_hint)
    extras: Sequence[Any] = []
    valid_values: Sequence[Any] | None = None
    if origin is Annotated:
        base, *extras = get_args(type_hint)
        LOGGER.debug('extras: %r', extras)
    else:
        base = type_hint
    if get_origin(base) is Literal:
        valid_values = get_literal_values(base)
        LOGGER.debug('Literal: %r (values: %r)', base, valid_values)
    field_info = next((e for e in extras if isinstance(e, FieldInfo)), None)
    if field_info is not None:
        return base, parse_pydantic_field_info(field_info, valid_values=valid_values)
    return type_hint, ParsedFieldInfo(valid_values=valid_values)


def get_gradio_component(type_hint) -> gr.Component | None:
    origin = get_origin(type_hint)
    if origin is Annotated:
        _base, *extras = get_args(type_hint)
        LOGGER.debug('extras: %r', extras)
        component = next((e for e in extras if isinstance(e, gr.Component)), None)
        return component
    return None


class FnGradio:
    def get_component(  # pylint: disable=too-many-return-statements
        self,
        type_hint: Any,
        default_value: Any | None = None
    ) -> gr.Component:
        gradio_component = get_gradio_component(type_hint)
        if gradio_component is not None:
            return gradio_component
        parsed_type_hint, field_info = parse_pydantic_field(type_hint)
        LOGGER.debug('field_info: %r', field_info)
        label = field_info.title
        if parsed_type_hint is bool:
            return gr.Checkbox(value=bool(default_value), label=label)
        if parsed_type_hint is int and field_info is not None:
            if field_info.ge is not None and field_info.le is not None:
                return gr.Slider(
                    minimum=cast(float, field_info.ge),
                    maximum=cast(float, field_info.le),
                    value=default_value,
                    label=label,
                    step=1
                )
        if parsed_type_hint is int:
            return gr.Number(precision=0, value=default_value, label=label)
        if parsed_type_hint is float:
            return gr.Number(value=default_value, label=label)
        if parsed_type_hint is str:
            return gr.Textbox(value=default_value, label=label)
        if field_info and field_info.valid_values:
            return gr.Dropdown(
                value=default_value,
                label=label,
                choices=list(field_info.valid_values)
            )
        raise UnsupportedTypeError(type_hint)

    def interface(
        self,
        *,
        api_name: str | None = None,
        **kwargs
    ) -> Callable[[Callable], gr.Interface]:
        def wrapper(fn: Callable) -> gr.Interface:
            hints = get_type_hints(fn, include_extras=True)
            sig = inspect.signature(fn)
            inputs = [
                self.get_component(
                    type_hint=hints.get(name),
                    default_value=param.default if param.default is not param.empty else None
                )
                for name, param in sig.parameters.items()
            ]
            outputs = [
                self.get_component(hints.get('return'))
            ]
            return gr.Interface(
                fn=fn,
                api_name=api_name or fn.__name__,
                inputs=inputs,
                outputs=outputs,
                **kwargs
            )
        return wrapper


DEFAULT_FNGRADIO = FnGradio()

interface = DEFAULT_FNGRADIO.interface



from typing import Annotated, Literal
import gradio as gr
from pydantic import Field
import pytest

from fngradio.decorators import FnGradio, UnsupportedTypeError, get_literal_values


@pytest.fixture(name='fngr')
def _fngr() -> FnGradio:
    return FnGradio()


class TestGetLiteralValues:
    def test_should_return_values(self):
        assert get_literal_values(Literal['a', 'b']) == ('a', 'b')


class TestFnGradio:
    class TestGetComponent:
        def test_should_use_gradio_component_if_present(self, fngr: FnGradio):
            component = fngr.get_component(Annotated[str, gr.TextArea()])
            assert isinstance(component, gr.TextArea)

        def test_should_map_str_to_textbox(self, fngr: FnGradio):
            component = fngr.get_component(str)
            assert isinstance(component, gr.Textbox)

        def test_should_map_str_with_label_to_textbox(self, fngr: FnGradio):
            component = fngr.get_component(
                Annotated[str, Field(title='Label 1')],
                default_value='Default 1'
            )
            assert isinstance(component, gr.Textbox)
            assert component.value == 'Default 1'
            assert component.label == 'Label 1'

        def test_should_map_int_to_number_with_zero_precision(self, fngr: FnGradio):
            component = fngr.get_component(int, default_value=10)
            assert isinstance(component, gr.Number)
            assert component.precision == 0
            assert component.value == 10

        def test_should_map_float_to_number_with_default_precision(self, fngr: FnGradio):
            component = fngr.get_component(float, default_value=10.5)
            assert isinstance(component, gr.Number)
            assert component.precision != 0
            assert component.value == 10.5

        def test_should_map_bool_to_checkbox_without_default(self, fngr: FnGradio):
            component = fngr.get_component(
                type_hint=bool,
                default_value=None
            )
            assert isinstance(component, gr.Checkbox)
            assert component.value is False

        def test_should_map_bool_to_checkbox_with_true_default(self, fngr: FnGradio):
            component = fngr.get_component(
                type_hint=bool,
                default_value=True
            )
            assert isinstance(component, gr.Checkbox)
            assert component.value is True

        def test_should_map_literal_to_dropdown(self, fngr: FnGradio):
            component = fngr.get_component(
                type_hint=Literal['value_1', 'value_2', 'value_3'],
                default_value='value_2'
            )
            assert isinstance(component, gr.Dropdown)
            assert component.choices == [
                ('value_1', 'value_1'),
                ('value_2', 'value_2'),
                ('value_3', 'value_3')
            ]
            assert component.value == 'value_2'

        def test_should_map_annotated_literal_to_dropdown(self, fngr: FnGradio):
            component = fngr.get_component(
                type_hint=Annotated[
                    Literal['value_1', 'value_2', 'value_3'],
                    Field(title='Label 1')
                ],
                default_value='value_2'
            )
            assert isinstance(component, gr.Dropdown)
            assert component.choices == [
                ('value_1', 'value_1'),
                ('value_2', 'value_2'),
                ('value_3', 'value_3')
            ]
            assert component.value == 'value_2'
            assert component.label == 'Label 1'

        def test_should_map_int_field_with_range_to_slider(self, fngr: FnGradio):
            component = fngr.get_component(
                type_hint=Annotated[int, Field(ge=10, le=100, title='Label 1')],
                default_value=50
            )
            assert isinstance(component, gr.Slider)
            assert component.minimum == 10
            assert component.maximum == 100
            assert component.value == 50
            assert component.step == 1
            assert component.label == 'Label 1'

    class TestInterface:
        def test_should_fail_without_type_hints(
            self,
            fngr: FnGradio
        ):
            def fn(s):
                return s.upper()

            with pytest.raises(UnsupportedTypeError):
                fngr.interface()(fn)

        def test_should_create_gradio_interface_with_simple_type(
            self,
            fngr: FnGradio
        ):
            @fngr.interface()
            def fn(s: str) -> str:
                return s.upper()

            assert fn.input_components and len(fn.input_components) == 1
            assert isinstance(fn.input_components[0], gr.Textbox)
            assert len(fn.output_components) == 1
            assert isinstance(fn.output_components[0], gr.Textbox)

        def test_should_create_gradio_interface_with_slider(
            self,
            fngr: FnGradio
        ):
            @fngr.interface()
            def fn(n: Annotated[int, Field(ge=10, le=100)] = 50) -> int:
                return n

            assert fn.input_components and len(fn.input_components) == 1
            input_component = fn.input_components[0]
            assert isinstance(input_component, gr.Slider)
            assert input_component.minimum == 10
            assert input_component.maximum == 100
            assert input_component.value == 50

        def test_should_set_api_name_to_function_name(
            self,
            fngr: FnGradio
        ):
            @fngr.interface()
            def fn(s: str) -> str:
                return s.upper()

            assert fn.api_name == 'fn'

        def test_should_allow_overriding_api_name(
            self,
            fngr: FnGradio
        ):
            @fngr.interface(api_name='fn_1')
            def fn(s: str) -> str:
                return s.upper()

            assert fn.api_name == 'fn_1'



from typing import Sequence
import gradio as gr


def tabbed_interface(
    interface_list: Sequence[gr.Blocks],
    tab_names: Sequence[str] | None = None,
    **kwargs
) -> gr.TabbedInterface:
    if not tab_names:
        tab_names = [
            (
                interface.api_name
                if isinstance(interface, gr.Interface) and interface.api_name
                else f'Tab {1 + index}'
            )
            for index, interface in enumerate(interface_list)
        ]
    return gr.TabbedInterface(
        interface_list=interface_list,
        tab_names=list(tab_names),
        **kwargs
    )
